# -*- coding: utf-8 -*-
# Copyright (C) 2020  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

"""Package xbtree provides utilities for inspecting/manipulating internal
structure of integer-keyed BTrees.

It is primarily used to verify ΔBTail in wcfs.

- `Tree` represents a tree node.
- `Bucket` represents a bucket node.
- `StructureOf` returns internal structure of a BTree represented as Tree and
  Bucket nodes.
- `Restructure` XXX

- `AllStructs` XXX


Topology encoding
-----------------

Topology encoding provides way to represent structure of a Tree as path-like string.

TopoEncode converts Tree into its topology-encoded representation, while
TopoDecode decodes topology-encoded string back into Tree.

The following example illustrates topology encoding represented by string
"T3/T-T/B1-T5/B-B7,8,9":

      [ 3 ]             T3/         represents Tree([3])
       / \
     [ ] [ ]            T-T/        represents two empty Tree([])
      ↓   ↓
     |1|[ 5 ]           B1-T5/      represent Bucket(1) and Tree([5])
         / \
        || |7|8|9|      B-B7,8,9    represents empty Bucket() and Bucket(7,8,9)


Topology encoding specification:

A Tree is encoded by level-order traversal, delimiting layers with "/".
Inside a layer Tree and Bucket nodes are signalled as

    "T<keys>"       ; Tree
    "B<keys>"       ; Bucket

Keys are represented as ","-delimited list of integers. For example Tree
with [1,3,5] keys is represented as

    "T1,3,5"        ; Tree([1,3,5])

Nodes inside one layer are delimited with "-". For example a layer consisting
of an empty Tree, a Tree with [1,3] keys, and Bucket with [4,5] keys is
represented as

    "T-T1,3-B4,5"   ; layer with Tree([]), Tree([1,3]) and Bucket([4,5])

A layer consists of nodes that are followed by node-node links from upper layer
in left-to-right order.
"""

from __future__ import print_function, absolute_import

from BTrees import check as bcheck
from golang import func, panic
from golang.gcompat import qq
import itertools
import re
inf = float('inf')


# Tree represents a tree node.
class Tree(object):
    #   .keyv       () of keys
    #   .children   () of children  len(.children) == len(.keyv) + 1
    def __init__(t, keyv, *children):
        assert len(children) == len(keyv) + 1, (keyv, children)
        _assertIncv(keyv)
        if len(children) > 0:
            # assert all children are of the same type
            childtypes = set([type(_) for _ in children])
            if len(childtypes) != 1:
                panic("children are of distinct types: %s" % (childtypes,))

            # assert type(child) is Tree | Bucket
            childtype = childtypes.pop()
            assert childtype in (Tree, Bucket), childtype

            # assert children keys are consistent
            v = [-inf] + keyv + [+inf]
            for i, (klo, khi) in enumerate(zip(v[:-1], v[1:])): # (klo, khi) = [] of (k_i, k_{i+1})
                child = children[i]
                for k in children[i].keyv:
                    if not (klo <= k < khi):
                        panic("children[%d] key %d is outside key range [%s, %s)" % (i, k, klo, khi))

        t.keyv     = tuple(keyv)
        t.children = tuple(children)

    def __hash__(t):
        return hash(t.keyv) ^ hash(t.children)
    def __ne__(a, b):
        return not (a == b)
    def __eq__(a, b):
        if not isinstance(b, Tree):
            return False
        return (a.keyv == b.keyv) and (a.children == b.children)

    def __str__(t):
        s = "T([" + ",".join(['%s' % _ for _ in t.keyv]) + "]"
        for ch in t.children:
            s += ",\n"
            s += _indent(' '*4, str(ch))
        s += ")"
        return s

    __repr__ = __str__
    #def __repr__(t):
    #    return 'R'+str(t)

    # TODO def graphviz() -> str for graphviz

# Bucket represents a bucket node.
class Bucket(object):
    #   .keyv   () of keys
    def __init__(b, *keyv):     # XXX *keyv -> keyv ?
        _assertIncv(keyv)
        b.keyv = tuple(keyv)

    def __hash__(b):
        return hash(b.keyv)
    def __ne__(a, b):
        return not (a == b)
    def __eq__(a, b):
        if not isinstance(b, Bucket):
            return False
        return a.keyv == b.keyv

    def __str__(b):
        return "B(" + ','.join(['%s' % _ for _ in b.keyv]) + ")"

    __repr__ = __str__


# StructureOf returns internal structure of a BTree.
#
# The structure is represented as Tree and Bucket nodes.
def StructureOf(node):
    typ = type(node)
    is_tree   = ("Tree"   in typ.__name__)
    is_set    = ("Set"    in typ.__name__)
    is_bucket = (("Bucket" in typ.__name__) or re.match("..Set", typ.__name__))
    is_map    = (not is_set)

    if is_bucket:
        keys, _ = bcheck.crack_bucket(node, is_map)
        return Bucket(*keys)

    if is_tree:
        kind, keys, children = bcheck.crack_btree(node, is_map)
        if kind == bcheck.BTREE_EMPTY:
            return Tree([], Bucket())

        if kind == bcheck.BTREE_ONE:
            b = node._bucket_type()
            b.__setstate__(keys)    # it is keys+values for BTREE_ONE case
            return Tree([], StructureOf(b))

        if kind == bcheck.BTREE_NORMAL:
            return Tree(keys, *[StructureOf(_) for _ in children])

        panic("bad tree kind %r" % kind)

    panic("unknown node type %r" % typ)


# TopoEncode returns topology encoding for internal structure of the tree.
#
# See top-level docstring for description of topology encoding.
def TopoEncode(tree):
    assert isinstance(tree, Tree)
    topo = ''

    # breadth-first traversal of the tree with '/' injected in between layers
    for nodev in _walkBFS(tree):
        if len(topo) != 0:
            topo += '/'
        tnodev = []
        for node in nodev:
            assert isinstance(node, (Tree, Bucket))
            tnode = ('T' if isinstance(node, Tree) else 'B') + \
                        (','.join(['%d' % _ for _ in node.keyv]))
            tnodev.append(tnode)
        topo += '-'.join(tnodev)


    """
    queue  = [tree, '/']
    while 1:
        x = queue.pop(0)
        if x == '/':
            if len(queue) == 0:
                break   # it was last '/'
            topo += '/'
            queue.append('/')   # add / after all queued nodes of next layer
            continue

        node = x
        assert isinstance(node, (Tree, Bucket))
        tnode = ('T' if isinstance(node, Tree) else 'B') + \
                    (','.join(['%d' % _ for _ in node.keyv]))
        if len(topo) != 0 and topo[-1] != '/':
            topo += '-'
        topo += tnode

        if isinstance(node, Tree):
            queue += node.children
    """

    if 1: # make sure that every topology we emit, can be loaded back
        t2 = TopoDecode(topo)
        if t2 != tree:
            panic("BUG: TopoEncode: D(E(·)) != identity\n·      = %s\n D(E(·) = %s" % (tree, t2))
    return topo


# _walkBFS walks tree in breadth-first order layer by layer.
def _walkBFS(tree): # i[] of [](of nodes on each level)
    assert isinstance(tree, Tree)
    currentv = []
    nextq    = [tree]
    while len(nextq) > 0:
        yield tuple(nextq)
        currentq = nextq
        nextq    = []
        while len(currentq) > 0:
            node = currentq.pop(0)
            assert isinstance(node, (Tree, Bucket))
            if isinstance(node, Tree):
                nextq += node.children



# TopoDecode decodes topology-encoded text into Tree structure.
#
# See top-level docstring for description of topology encoding.
class TopoDecodeError(Exception):
    pass
def TopoDecode(text):
    levelv = text.split('/')    # T3/T-T/B1-T5/B-B7,8,9  -> T3 T-T B1-T5 B-B7,8,9

    # build nodes from bottom-up
    currentv = [] # of nodes on current level (that we are building)
    bottomq  = [] # of nodes below current level that we are building
                  # shrinks as fifo as nodes added to currentv link to bottom
    while len(levelv) > 0:
        level  = levelv.pop()       # e.g. B1-T5
        tnodev = level.split('-')   # e.g. B1 T5
        bottomq  = currentv
        currentv = []
        for tnode in tnodev:
            if   tnode[:1] == 'T':
                typ = Tree
            elif tnode[:1] == 'B':
                typ = Bucket
            else:
                raise TopoDecodeError("incorrect node %s: unknown prefix" % qq(tnode))
            tkeys = tnode[1:]    # e.g. '7,8,9' or ''
            if tkeys == '':
                keyv = []
            else:
                tkeyv = tkeys.split(',')    # e.g. 7 8 9
                keyv  = [int(_) for _ in tkeyv]
            if typ is Bucket:
                node = Bucket(*keyv)
            else:
                # Tree
                nchild = len(keyv) + 1
                if len(bottomq) < nchild:
                    raise TopoDecodeError(
                        "node %s at level %d: next level does not have enough children to link to" %
                        (qq(tnode), len(levelv)+1))
                children = bottomq[:nchild]
                bottomq  = bottomq[nchild:]
                node = Tree(keyv, *children)

            currentv.append(node)

        if len(bottomq) != 0:
            raise TopoDecodeError("level %d does not link to all nodes in the next level" %
                                  len(levelv+1))

    if len(currentv) != 1:
        raise TopoDecodeError("first level has %d entries; must be 1" % len(currentv))

    root = currentv[0]
    return root


# Restructure reorganizes BTree instance (not Tree) according to new structure.
def Restructure(tree, newStructure):
    assert istree(tree)
    assert isinstance(newStructure, Tree)
    1/0

# AllStructs generates subset of all possible BTree structures for BTrees with
# specified keys and btree depth up-to maxdepth. Each tree node is split by
# up-to maxsplit points.
def AllStructs(keys, maxdepth, maxsplit): # -> i[] of Tree
    assert isinstance(maxdepth, int); assert maxdepth >= 0
    assert isinstance(maxsplit, int); assert maxsplit >= 0
    ks = set(keys)
    for k in keys:
        assert isinstance(k, int)
        assert k in ks  # no duplicates
        ks.remove(k)
    keyv = list(keys)
    keyv.sort()

    # initial [lo, hi) covering keys and such that split points will be there withing +-1 of min/max key
    #klo = -inf
    #khi = +inf
    if len(keyv) > 0:
        klo = keyv[0]  - 1 - 1
        khi = keyv[-1] + 1 + 1   # hi is ")", not "]"
    else:
        # XXX ok? (should be -inf,+inf)
        klo = 0
        khi = 0

    for tree in _allStructs(klo, khi, keyv, maxdepth, maxsplit):
        yield tree


def _allStructs(klo, khi, keyv, maxdepth, maxsplit):
    assert klo <= khi
    _assertIncv(keyv)
    if len(keyv) > 0:
        assert klo <= keyv[0]
        assert        keyv[-1] < khi

    #print('_allStructs [%s, %s)  keyv: %r,  maxdepth=%d, maxsplit=%d' %
    #        (klo, khi, keyv, maxdepth, maxsplit))

    for nsplit in range(0, maxsplit+1):
        for ksplitv in _iterSplitByN(klo, khi, nsplit):
            # ksplitv = [klo, s1, s2, ..., sN, khi]
            #print('ksplitv: %r' % ksplitv)

            # emit Tree -> Buckets
            children = []
            for (xlo, xhi) in zip(ksplitv[:-1], ksplitv[1:]): # (klo, s1), (s1, s2), ..., (sN, khi)
                children.append(Bucket(*_keyvSliceBy(keyv, xlo, xhi)))
            yield Tree(ksplitv[1:-1], *children) # (s1, s2, ..., sN)

            # emit Tree -> Trees -> ...
            if maxdepth == 0:
                continue
            ichildrenv = [] # of _allStructs for each child link
            for (xlo, xhi) in zip(ksplitv[:-1], ksplitv[1:]): # (klo, s1), (s1, s2), ..., (sN, khi)
                ichildrenv.append( _allStructs(
                    xlo, xhi, _keyvSliceBy(keyv, xlo, xhi), maxdepth - 1, maxsplit))

            for children in itertools.product(*ichildrenv):
                yield Tree(ksplitv[1:-1], *children) # (s1, s2, ..., sN)


# _keyvSliceBy returns [] of keys from keyv : k ∈ [klo, khi)
def _keyvSliceBy(keyv, klo, khi):
    assert klo <= khi
    _assertIncv(keyv)
    return list([k for k in keyv if (klo <= k < khi)])

# _iterSplitByN iterates through all nsplit splitting of [lo, hi) range.
# hi > lo
# XXX nsplit > ...
# lo < si < hi
# si <  s_{i+1}
#
# XXX remove lo and hi from the output?
def _iterSplitByN(lo, hi, nsplit): # -> i[] of [lo, s1, s2, ..., sn, hi)
    assert lo <= hi
    assert nsplit >= 0

    if nsplit == 0:
        yield [lo, hi]
        return

    for s in range(lo+1, hi):   # [lo+1, hi-1]
        for tail in _iterSplitByN(s, hi, nsplit-1):
            yield [lo] + tail


# ---- misc ----

# _indent returns text with each line of it indented with prefix.
def _indent(prefix, text): # -> text
    textv = text.split('\n')
    textv = [prefix+_ for _ in textv]
    text  = '\n'.join(textv)
    return text

# _assertIncv asserts that values of vector v are strictly ↑
def _assertIncv(v):
    prev = -inf
    for i in range(len(v)):
        if not (v[i] > prev):
            panic("assert incv: [%d] not ↑: %s -> %s" % (i, v[i], prev))
        prev = v[i]





# ----------------------------------------

import sys

# graphviz returns tree graph representation in dot language.
@func(Tree)
def graphviz(t):
    assert isinstance(t, Tree)

    symtab = {} # node -> name

    outv = []
    def emit(line):
        outv.append(line)

    emit('digraph {')
    emit('  splines=false')

    for (level, nodev) in enumerate(_walkBFS(t)):
        for (i, node) in enumerate(nodev):
            assert isinstance(node, (Tree,Bucket))

            # register node insymtab
            assert node not in symtab
            kind = ('T' if isinstance(node, Tree) else 'B')
            symtab[node] = '%s%d:%d' % (kind, level, i)

            # emit node itself
            # approach based on https://github.com/Felerius/btree-generator
            emit('  %s' % qq(symtab[node]))
            emit('  [')
            emit('    shape = none')
            emit('    label = <<table border="1" cellborder="0" cellspacing="0">')
            emit('                <tr>')
            for (j,key) in enumerate(node.keyv):
                if kind == 'T':
                    emit('                  <td port="con%d"></td>' % j)
                if 1:
                    emit('                  <td port="key%d">%d</td>' % (j,key))
            emit('                  <td port="con%d">%s</td>' % (len(node.keyv), \
                                            ('    ' if len(node.keyv) == 0 else '')))

            emit('                </tr>')
            emit('             </table>>')
            emit('  ]')

    # second pass: emit links + node ranks
    for nodev in _walkBFS(t):
        # same rank for nodes on the same level
        emit('')
        emit('  {rank=same; %s}' % ' '.join([qq(symtab[_]) for _ in nodev]))
        # links
        for node in nodev:
            if not isinstance(node, Tree):
                continue
            for (j,child) in enumerate(node.children):
                emit('%s:"con%d" -> %s' % (qq(symtab[node]), j, qq(symtab[child])))

    emit('}')
    emit('')
    return '\n'.join(outv)


# topoview display topologies provided in argv.
def topoview(argv):
    if len(argv) == 0:
        raise RuntimeError('E: empty argument')

    topov = [TopoDecode(_) for _ in argv]

    import graphviz as gv
    #g = gv.Digraph(comment='AAA')
    #print(g.render(format='pdf', view=True))
    g = gv.Source(topov[0].graphviz())
    #g.view()
    g.render('x', format='svg')


if __name__ == '__main__':
    topoview(sys.argv[1:])
