# -*- coding: utf-8 -*-
# Copyright (C) 2020  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

from __future__ import print_function, absolute_import


from wendelin.wcfs.internal import xbtree
from BTrees.LOBTree import LOBTree
from BTrees.tests import testBTrees

def test_structureOf():
    T = xbtree.Tree
    B = xbtree.Bucket

    # empty tree
    t = LOBTree()
    assert xbtree.StructureOf(t) == T([], B())

    # tree with 1 k->v
    t = LOBTree()
    t[10] = 'hello'
    assert xbtree.StructureOf(t) == T([], B(10))

    # known degenerate topology, see:
    # https://github.com/zopefoundation/ZODB/commit/6cd24e99f89b
    # https://github.com/zopefoundation/BTrees/blob/4.7.2-1-g078ba60/BTrees/tests/testBTrees.py#L20
    t, keys = testBTrees.DegenerateBTree("testBasicOps")._build_degenerate_tree()
    assert keys == [1, 3, 5, 7, 11]
    assert xbtree.StructureOf(t) == T([4],
                                      T([2],
                                           T([], B(1)), T([], B(3))),
                                      T([],
                                           T([6, 10],
                                               T([], T([], B(5))),
                                               T([], B(7)),
                                               T([], B(11))) ))

def test_topoEncoding():
    T = xbtree.Tree
    B = xbtree.Bucket
    def X(tree):
        topo = xbtree.TopoEncode(tree)
        t2   = xbtree.TopoDecode(topo)
        assert t2 == tree
        return topo

    assert X(T([], B()))            == 'T/B'
    assert X(T([], B(1)))           == 'T/B1'
    assert X(T([], B(1,3)))         == 'T/B1,3'
    assert X(T([], T([], B())))     == 'T/T/B'

    assert X(T([3],
               T([], B(1)),
               T([],
                 T([5], B(), B(7,8,9)))))   == "T3/T-T/B1-T5/B-B7,8,9"

    # degenerate btree from ZODB
    assert X(T([4],
               T([2],
                    T([], B(1)), T([], B(3))),
               T([],
                    T([6, 10],
                        T([], T([], B(5))),
                        T([], B(7)),
                        T([], B(11))) )))   == "T4/T2-T/T-T-T6,10/B1-B3-T-T-T/T-B7-B11/B5"


    for tree in xbtree.AllStructs([1,3,7,8], 1,2):
        t2 = xbtree.TopoDecode(xbtree.TopoEncode(tree))
        assert t2 == tree


def test_allStructs():
    T = xbtree.Tree
    B = xbtree.Bucket
    def X(keys, maxdepth, maxsplit):
        return list(xbtree.AllStructs(keys, maxdepth, maxsplit))

    assert X([],    0, 0) == [ T([], B()) ]
    assert X([1],   0, 0) == [ T([], B(1)) ]
    assert X([1,3], 0, 0) == [ T([], B(1,3)) ]

    assert X([],    0, 1) == [ T([], B()) ] # nothing to split (?)
    #assert X([],    0, 1) == [ T([], B()), T([0], B(), B()) ]


    assert X([],    1, 0) == [ T([], B()),
                               T([],
                                 T([], B())) ]

    assert X([],    2, 0) == [ T([], B()),
                               T([],
                                 T([], B())),
                               T([],
                                 T([],
                                   T([], B()))) ]


    assert X([1,3], 0, 0) == [ T([], B(1,3)) ]
    assert X([1,3], 0, 1) == [ T([], B(1,3)),
                               # nsplit=1
                               T([0], B(), B(1,3)),
                               T([1], B(), B(1,3)),
                               T([2], B(1), B(3)),
                               T([3], B(1), B(3)),
                               T([4], B(1,3), B()),
                             ]

    assert X([1,3], 1, 0) == [ T([], B(1,3)),
                               # depth=1
                               T([],
                                 T([], B(1,3)))
                             ]

    assert X([1,3], 1, 1) == [
                               # T/
                               T([], B(1,3)),           # nsplit=0
                               T([],                    # nsplit=0,0
                                 T([], B(1,3))),
                               T([],                    # nsplit=0,1
                                 T([0], B(), B(1,3))),
                               T([],
                                 T([1], B(), B(1,3))),
                               T([],
                                 T([2], B(1), B(3))),
                               T([],
                                 T([3], B(1), B(3))),
                               T([],
                                 T([4], B(1,3), B())),

                               # T0/
                               T([0], B(), B(1,3)),     # nsplit=1
                               T([0],                   # nsplit=1,(0,0)
                                 T([], B()),
                                 T([], B(1,3))),
                               T([0],                   # nsplit=1,(0,1)
                                 T([],  B()),
                                 T([1], B(),  B(1,3))),
                               T([0],
                                 T([],  B()),
                                 T([2], B(1), B(3))),
                               T([0],
                                 T([],  B()),
                                 T([3], B(1), B(3))),
                               T([0],
                                 T([],  B()),
                                 T([4], B(1,3), B())),
                                                        # nsplit=1,(1,*) -> ø

                               # T1/
                               T([1], B(), B(1,3)),     # nsplit=1
                               T([1],                   # nsplit=1,(0,0)
                                 T([], B()),
                                 T([], B(1,3))),
                               T([1],
                                 T([],  B()),           # nsplit=1,(0,1)
                                 T([2], B(1), B(3))),
                               T([1],
                                 T([],  B()),
                                 T([3], B(1), B(3))),
                               T([1],
                                 T([],  B()),
                                 T([4], B(1,3), B())),
                               T([1],                   # nsplit=1,(1,0)
                                 T([0], B(),  B()),
                                 T([],  B(1,3))),
                               T([1],                   # nsplit=1,(1,1)
                                 T([0], B(),  B()),
                                 T([2], B(1), B(3))),
                               T([1],
                                 T([0], B(),  B()),
                                 T([3], B(1), B(3))),
                               T([1],
                                 T([0], B(),  B()),
                                 T([4], B(1,3), B())),

                               # T2/
                               T([2], B(1), B(3)),      # nsplit=1
                               T([2],                   # nsplit=1,(0,0)
                                 T([], B(1)),
                                 T([], B(3))),
                               T([2],                   # nsplit=1,(0,1)
                                 T([],  B(1)),
                                 T([3], B(),  B(3))),
                               T([2],
                                 T([],  B(1)),
                                 T([4], B(3), B())),
                               T([2],                   # nsplit=1,(1,0)
                                 T([0], B(),  B(1)),
                                 T([],  B(3))),
                               T([2],                   # nsplit=1,(1,1)
                                 T([0], B(),  B(1)),
                                 T([3], B(),  B(3))),
                               T([2],
                                 T([0], B(),  B(1)),
                                 T([4], B(3), B())),
                               T([2],                   # nsplit=1,(1,0)
                                 T([1], B(),  B(1)),
                                 T([],  B(3))),
                               T([2],                   # nsplit=1,(1,1)
                                 T([1], B(),  B(1)),
                                 T([3], B(),  B(3))),
                               T([2],
                                 T([1], B(),  B(1)),
                                 T([4], B(3), B())),

                               # T3/
                               T([3], B(1), B(3)),      # nsplit=1
                               T([3],                   # nsplit=1,(0,0)
                                 T([], B(1)),
                                 T([], B(3))),
                               T([3],                   # nsplit=1,(0,1)
                                 T([],  B(1)),
                                 T([4], B(3), B())),
                               T([3],                   # nsplit=1,(1,0)
                                 T([0], B(),  B(1)),
                                 T([],  B(3))),
                               T([3],                   # nsplit=1,(1,1)
                                 T([0], B(),  B(1)),
                                 T([4], B(3), B())),
                               T([3],                   # nsplit=1,(1,0)
                                 T([1], B(),  B(1)),
                                 T([],  B(3))),
                               T([3],                   # nsplit=1,(1,1)
                                 T([1], B(),  B(1)),
                                 T([4], B(3), B())),
                               T([3],                   # nsplit=1,(1,0)
                                 T([2], B(1), B()),
                                 T([],  B(3))),
                               T([3],                   # nsplit=1,(1,1)
                                 T([2], B(1), B()),
                                 T([4], B(3), B())),

                               # T4/
                               T([4], B(1,3), B()),     # nsplit=1
                               T([4],                   # nsplit=1,(0,0)
                                 T([], B(1,3)),
                                 T([], B())),
                                                        # nsplit=1,(0,1) -> ø
                               T([4],                   # nsplit=1,(1,0)
                                 T([0], B(),  B(1,3)),
                                 T([],  B())),
                               T([4],
                                 T([1], B(),  B(1,3)),
                                 T([],  B())),
                               T([4],
                                 T([2], B(1), B(3)),
                                 T([],  B())),
                               T([4],
                                 T([3], B(1), B(3)),
                                 T([],  B())),
                                                        # nsplit=1,(1,1) -> ø
                              ]

    # XXX tests for maxsplit=2 / maxdepth=2 ?
    # (if yes -> compare with set, not exact order, and use topoencoding on the right)



def test_keyvSliceBy():
    X = xbtree._keyvSliceBy
    assert X([],  0,0)  == []
    assert X([1], 0,0)  == []
    assert X([1], 0,1)  == []
    assert X([1], 1,1)  == []
    assert X([1], 1,2)  == [1]

    assert X([1,3,5,10,17], 3,10) == [3,5]

def test_iterSplitByN():
    def X(lo, hi, nsplit):
        return tuple(xbtree._iterSplitByN(lo, hi, nsplit))


    assert X(0,0, 0)    == ( [0,0], )
    assert X(0,0, 1)    == ()
    assert X(0,1, 0)    == ( [0,1], )
    assert X(0,1, 1)    == ()

    assert X(0,2, 0)    == ( [0,2], )
    assert X(0,2, 1)    == ( [0,1,2], )
    assert X(0,2, 2)    == ()

    assert X(0,3, 0)    == ( [0,3], )
    assert X(0,3, 1)    == ( [0,1,3], [0,2,3] )
    assert X(0,3, 2)    == ( [0,1,2,3], )
    assert X(0,3, 3)    == ()

    assert X(0,4, 0)    == ( [0,4], )
    assert X(0,4, 1)    == ( [0,1,4], [0,2,4], [0,3,4] )
    assert X(0,4, 2)    == ( [0,1,2,4], [0,1,3,4], [0,2,3,4] )
    assert X(0,4, 3)    == ( [0,1,2,3,4], )
    assert X(0,4, 4)    == ()

    assert X(0,5, 0)    == ( [0,5], )
    assert X(0,5, 1)    == ( [0,1,5], [0,2,5], [0,3,5], [0,4,5] )
    assert X(0,5, 2)    == ( [0,1,2,5], [0,1,3,5], [0,1,4,5], [0,2,3,5], [0,2,4,5], [0,3,4,5] )
    assert X(0,5, 3)    == ( [0,1,2,3,5], [0,1,2,4,5], [0,1,3,4,5], [0,2,3,4,5] )
    assert X(0,5, 4)    == ( [0,1,2,3,4,5], )
    assert X(0,5, 5)    == ()
